@@ -18,7 +18,7 @@
 
 ## You should have received a copy of the GNU General Public License
 ## along with this program; if not, write to the Free Software
-## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 
 import dbus.service
 import gobject
@@ -26,8 +26,13 @@ import sys
 
 from debug import *
 import asyncconn
+import config
+import cups
+import cupshelpers
+import dnssdresolve
 import jobviewer
 import newprinter
+import PhysicalDevice
 import ppdcache
 import printerproperties
 
@@ -38,6 +43,9 @@ CONFIG_NEWPRINTERDIALOG_IFACE=CONFIG_IFACE + ".NewPrinterDialog"
 CONFIG_PRINTERPROPERTIESDIALOG_IFACE=CONFIG_IFACE + ".PrinterPropertiesDialog"
 CONFIG_JOBVIEWER_IFACE=CONFIG_IFACE + ".JobViewer"
 
+g_ppds = None
+g_killtimer = None
+
 class KillTimer:
     def __init__ (self, timeout=30, killfunc=None):
         self._timeout = timeout
@@ -63,18 +71,189 @@ class KillTimer:
         self._holds += 1
 
     def remove_hold (self):
-        self._holds -= 1
-        if self._holds == 0:
-            debugprint ("Kill timer started")
-            self._add_timeout ()
+        if self._holds > 0:
+            self._holds -= 1
+            if self._holds == 0:
+                debugprint ("Kill timer started")
+                self._add_timeout ()
 
     def alive (self):
         if self._holds == 0:
             gobject.source_remove (self._timer)
             self._add_timeout ()
 
+class FetchedPPDs(gobject.GObject):
+    __gsignals__ = {
+        'ready': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, []),
+        'error': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
+                  [gobject.TYPE_PYOBJECT])
+        }
+
+    def __init__ (self, cupsconn, language):
+        gobject.GObject.__init__ (self)
+        self._cupsconn = cupsconn
+        self._language = language
+        self._ppds = None
+
+    def is_ready (self):
+        return self._ppds != None
+
+    def get_ppds (self):
+        return self._ppds
+
+    def run (self):
+        debugprint ("FetchPPDs: running")
+        try:
+            self._cupsconn.getPPDs2 (reply_handler=self._cups_getppds_reply,
+                                     error_handler=self._cups_error)
+        except AttributeError:
+            # getPPDs2 requires pycups >= 1.9.52
+            self._cupsconn.getPPDs (reply_handler=self._cups_getppds_reply,
+                                    error_handler=self._cups_error)
+
+    def _cups_error (self, conn, exc):
+        debugprint ("FetchPPDs: error: %s" % repr (exc))
+        self.emit ('error', exc)
+
+    def _cups_getppds_reply (self, conn, result):
+        debugprint ("FetchPPDs: success")
+        self._ppds = cupshelpers.ppds.PPDs (result, language=self._language)
+        self.emit ('ready')
+
+gobject.type_register (FetchedPPDs)
+
+class GetBestDriversRequest:
+    def __init__ (self, device_id, device_make_and_model, device_uri,
+                  cupsconn, language, reply_handler, error_handler):
+        self.device_id = device_id
+        self.device_make_and_model = device_make_and_model
+        self.device_uri = device_uri
+        self.reply_handler = reply_handler
+        self.error_handler = error_handler
+        self._signals = []
+        debugprint ("+%s" % self)
+
+        g_killtimer.add_hold ()
+        global g_ppds
+        if g_ppds == None:
+            debugprint ("GetBestDrivers request: need to fetch PPDs")
+            g_ppds = FetchedPPDs (cupsconn, language)
+            self._signals.append (g_ppds.connect ('ready', self._ppds_ready))
+            self._signals.append (g_ppds.connect ('error', self._ppds_error))
+            g_ppds.run ()
+        else:
+            if g_ppds.is_ready ():
+                debugprint ("GetBestDrivers request: PPDs already fetched")
+                self._ppds_ready (g_ppds)
+            else:
+                debugprint ("GetBestDrivers request: waiting for PPDs")
+                self._signals.append (g_ppds.connect ('ready',
+                                                      self._ppds_ready))
+                self._signals.append (g_ppds.connect ('error',
+                                                      self._ppds_error))
+
+    def __del__ (self):
+        debugprint ("-%s" % self)
+
+    def _disconnect_signals (self):
+        for s in self._signals:
+            g_ppds.disconnect (s)
+
+    def _ppds_error (self, fetchedppds, exc):
+        self._disconnect_signals ()
+        self.error_handler (exc)
+
+    def _ppds_ready (self, fetchedppds):
+        self._disconnect_signals ()
+        ppds = fetchedppds.get_ppds ()
+
+        try:
+            if self.device_id:
+                id_dict = cupshelpers.parseDeviceID (self.device_id)
+            else:
+                id_dict = {}
+                (mfg,
+                 mdl) = cupshelpers.ppds.ppdMakeModelSplit (self.device_make_and_model)
+                id_dict["MFG"] = mfg
+                id_dict["MDL"] = mdl
+                id_dict["DES"] = ""
+                id_dict["CMD"] = []
+
+            fit = ppds.getPPDNamesFromDeviceID (id_dict["MFG"],
+                                                id_dict["MDL"],
+                                                id_dict["DES"],
+                                                id_dict["CMD"],
+                                                self.device_uri,
+                                                self.device_make_and_model)
+
+            ppdnamelist = ppds.orderPPDNamesByPreference (fit.keys ())
+
+            g_killtimer.remove_hold ()
+            self.reply_handler (map (lambda x: (x, fit[x]), ppdnamelist))
+        except Exception, e:
+            try:
+                g_killtimer.remove_hold ()
+            except:
+                pass
+
+            self.error_handler (e)
+
+class GroupPhysicalDevicesRequest:
+    def __init__ (self, devices, reply_handler, error_handler):
+        self.devices = devices
+        self.reply_handler = reply_handler
+        self.error_handler = error_handler
+        debugprint ("+%s" % self)
+
+        try:
+            g_killtimer.add_hold ()
+            need_resolving = {}
+            self.deviceobjs = {}
+            for device_uri, device_dict in self.devices.iteritems ():
+                deviceobj = cupshelpers.Device (device_uri, **device_dict)
+                self.deviceobjs[device_uri] = deviceobj
+                if device_uri.startswith ("dnssd://"):
+                    need_resolving[device_uri] = deviceobj
+
+            if len (need_resolving) > 0:
+                resolver = dnssdresolve.DNSSDHostNamesResolver (need_resolving)
+                resolver.resolve (reply_handler=self._group)
+            else:
+                self._group ()
+        except Exception, e:
+            g_killtimer.remove_hold ()
+            self.error_handler (e)
+
+    def __del__ (self):
+        debugprint ("-%s" % self)
+
+    def _group (self, resolved_devices=None):
+        # We can ignore resolved_devices because the actual objects
+        # (in self.devices) have been modified.
+        try:
+            self.physdevs = []
+            for device_uri, deviceobj in self.deviceobjs.iteritems ():
+                newphysicaldevice = PhysicalDevice.PhysicalDevice (deviceobj)
+                matched = False
+                try:
+                    i = self.physdevs.index (newphysicaldevice)
+                    self.physdevs[i].add_device (deviceobj)
+                except ValueError:
+                    self.physdevs.append (newphysicaldevice)
+
+            uris_by_phys = []
+            for physdev in self.physdevs:
+                uris_by_phys.append (map (lambda x: x.uri,
+                                          physdev.get_devices ()))
+
+            g_killtimer.remove_hold ()
+            self.reply_handler (uris_by_phys)
+        except Exception, e:
+            g_killtimer.remove_hold ()
+            self.error_handler (e)
+
 class ConfigPrintingNewPrinterDialog(dbus.service.Object):
-    def __init__ (self, bus, path, cupsconn, killtimer):
+    def __init__ (self, bus, path, cupsconn):
         bus_name = dbus.service.BusName (CONFIG_BUS, bus=bus)
         dbus.service.Object.__init__ (self, bus_name, path)
         self.dialog = newprinter.NewPrinterGUI()
@@ -87,7 +266,6 @@ class ConfigPrintingNewPrinterDialog(dbus.service.Object):
                                              self.on_printer_modified)]
         self._ppdcache = ppdcache.PPDCache ()
         self._cupsconn = cupsconn
-        self._killtimer = killtimer
         debugprint ("+%s" % self)
 
     def __del__ (self):
@@ -96,14 +274,14 @@ class ConfigPrintingNewPrinterDialog(dbus.service.Object):
     @dbus.service.method(dbus_interface=CONFIG_NEWPRINTERDIALOG_IFACE,
                          in_signature='uss', out_signature='')
     def NewPrinterFromDevice(self, xid, device_uri, device_id):
-        self._killtimer.add_hold ()
+        g_killtimer.add_hold ()
         self.dialog.init ('printer_with_uri', device_uri=device_uri,
                           devid=device_id, xid=xid)
 
     @dbus.service.method(dbus_interface=CONFIG_NEWPRINTERDIALOG_IFACE,
                          in_signature='uss', out_signature='')
     def ChangePPD(self, xid, name, device_id):
-        self._killtimer.add_hold ()
+        g_killtimer.add_hold ()
         self.xid = xid
         self.name = name
         self.device_id = device_id
@@ -139,13 +317,13 @@ class ConfigPrintingNewPrinterDialog(dbus.service.Object):
         pass
 
     def on_dialog_canceled(self, obj):
-        self._killtimer.remove_hold ()
+        g__killtimer.remove_hold ()
         self.DialogCanceled ()
         self.remove_handles ()
         self.remove_from_connection ()
 
     def on_printer_added(self, obj, name):
-        self._killtimer.remove_hold ()
+        g_killtimer.remove_hold ()
         self.PrinterAdded (name)
         self.remove_handles ()
         self.remove_from_connection ()
@@ -155,7 +333,7 @@ class ConfigPrintingNewPrinterDialog(dbus.service.Object):
             self.dialog.disconnect (handle)
 
 class ConfigPrintingPrinterPropertiesDialog(dbus.service.Object):
-    def __init__ (self, bus, path, xid, name, killtimer):
+    def __init__ (self, bus, path, xid, name):
         bus_name = dbus.service.BusName (CONFIG_BUS, bus=bus)
         dbus.service.Object.__init__ (self, bus_name=bus_name, object_path=path)
         self.dialog = printerproperties.PrinterPropertiesDialog ()
@@ -164,8 +342,7 @@ class ConfigPrintingPrinterPropertiesDialog(dbus.service.Object):
         self.closed_handle = handle
         self.dialog.show (name)
         self.dialog.dialog.set_modal (False)
-        self._killtimer = killtimer
-        killtimer.add_hold ()
+        g_killtimer.add_hold ()
 
     @dbus.service.method(dbus_interface=CONFIG_PRINTERPROPERTIESDIALOG_IFACE,
                          in_signature='', out_signature='')
@@ -180,22 +357,21 @@ class ConfigPrintingPrinterPropertiesDialog(dbus.service.Object):
 
     def on_dialog_closed (self, dialog):
         dialog.destroy ()
-        self._killtimer.remove_hold ()
+        g_killtimer.remove_hold ()
         self.Finished ()
         self.dialog.disconnect (self.closed_handle)
         self.remove_from_connection ()
 
 class ConfigPrintingJobApplet(dbus.service.Object):
-    def __init__ (self, bus, path, killtimer):
+    def __init__ (self, bus, path):
         bus_name = dbus.service.BusName (CONFIG_BUS, bus=bus)
         dbus.service.Object.__init__ (self, bus_name=bus_name, object_path=path)
         self.jobapplet = jobviewer.JobViewer(bus=dbus.SystemBus (),
                                              applet=True, my_jobs=True)
         handle = self.jobapplet.connect ('finished', self.on_jobapplet_finished)
         self.finished_handle = handle
-        self._killtimer = killtimer
         self.has_finished = False
-        killtimer.add_hold ()
+        g_killtimer.add_hold ()
         debugprint ("+%s" % self)
 
     def __del__ (self):
@@ -213,14 +389,13 @@ class ConfigPrintingJobApplet(dbus.service.Object):
 
     def on_jobapplet_finished (self, jobapplet):
         self.Finished ()
-        self._killtimer.remove_hold ()
+        g_killtimer.remove_hold ()
         self.has_finished = True
         self.jobapplet.disconnect (self.finished_handle)
         self.remove_from_connection ()
 
 class ConfigPrinting(dbus.service.Object):
-    def __init__ (self, killtimer):
-        self._killtimer = killtimer
+    def __init__ (self):
         self.bus = dbus.SessionBus ()
         bus_name = dbus.service.BusName (CONFIG_BUS, bus=self.bus)
         dbus.service.Object.__init__ (self, bus_name, CONFIG_PATH)
@@ -228,6 +403,8 @@ class ConfigPrinting(dbus.service.Object):
         self._pathn = 0
         self._jobapplet = None
         self._jobappletpath = None
+        self._ppds = None
+        self._language = locale.getlocale (locale.LC_MESSAGES)[0]
 
     def destroy (self):
         self._cupsconn.destroy ()
@@ -238,9 +415,8 @@ class ConfigPrinting(dbus.service.Object):
         self._pathn += 1
         path = "%s/NewPrinterDialog/%s" % (CONFIG_PATH, self._pathn)
         ConfigPrintingNewPrinterDialog (self.bus, path,
-                                        self._cupsconn,
-                                        killtimer=self._killtimer)
-        self._killtimer.alive ()
+                                        self._cupsconn)
+        g_killtimer.alive ()
         return path
 
     @dbus.service.method(dbus_interface=CONFIG_IFACE,
@@ -248,9 +424,8 @@ class ConfigPrinting(dbus.service.Object):
     def PrinterPropertiesDialog(self, xid, name):
         self._pathn += 1
         path = "%s/PrinterPropertiesDialog/%s" % (CONFIG_PATH, self._pathn)
-        ConfigPrintingPrinterPropertiesDialog (self.bus, path, xid, name,
-                                               killtimer=self._killtimer)
-        self._killtimer.alive ()
+        ConfigPrintingPrinterPropertiesDialog (self.bus, path, xid, name)
+        g_killtimer.alive ()
         return path
 
     @dbus.service.method(dbus_interface=CONFIG_IFACE,
@@ -259,12 +434,32 @@ class ConfigPrinting(dbus.service.Object):
        if self._jobapplet == None or self._jobapplet.has_finished:
             self._pathn += 1
             path = "%s/JobApplet/%s" % (CONFIG_PATH, self._pathn)
-            self._jobapplet = ConfigPrintingJobApplet (self.bus, path,
-                                                       self._killtimer)
+            self._jobapplet = ConfigPrintingJobApplet (self.bus, path)
             self._jobappletpath = path
 
        return self._jobappletpath
 
+    @dbus.service.method(dbus_interface=CONFIG_IFACE,
+                         in_signature='sss', out_signature='a(ss)',
+                         async_callbacks=('reply_handler', 'error_handler'))
+    def GetBestDrivers(self, device_id, device_make_and_model, device_uri,
+                   reply_handler, error_handler):
+        GetBestDriversRequest (device_id, device_make_and_model, device_uri,
+                               self._cupsconn, self._language[0],
+                               reply_handler, error_handler)
+
+    @dbus.service.method(dbus_interface=CONFIG_IFACE,
+                         in_signature='s', out_signature='as')
+    def MissingExecutables(self, ppd_filename):
+        ppd = cups.PPD (ppd_filename)
+        return cupshelpers.missingExecutables (ppd)
+
+    @dbus.service.method(dbus_interface=CONFIG_IFACE,
+                         in_signature='a{sa{ss}}', out_signature='aas',
+                         async_callbacks=('reply_handler', 'error_handler'))
+    def GroupPhysicalDevices(self, devices, reply_handler, error_handler):
+        GroupPhysicalDevicesRequest (devices, reply_handler, error_handler)
+
 def _client_demo ():
     # Client demo
     if len (sys.argv) > 2:
@@ -307,12 +502,15 @@ def _client_demo ():
 
 if __name__ == '__main__':
     import ppdippstr
+    import gettext
     import locale
     try:
         locale.setlocale (locale.LC_ALL, "")
     except:
         pass
 
+    gettext.textdomain (config.PACKAGE)
+    gettext.bindtextdomain (config.PACKAGE, config.localedir)
     ppdippstr.init ()
     gobject.threads_init ()
     from dbus.glib import DBusGMainLoop
@@ -332,7 +530,7 @@ if __name__ == '__main__':
 
     debugprint ("Service running...")
     loop = gobject.MainLoop ()
-    killtimer = KillTimer (killfunc=loop.quit)
-    cp = ConfigPrinting (killtimer=killtimer)
+    g_killtimer = KillTimer (killfunc=loop.quit)
+    cp = ConfigPrinting ()
     loop.run ()
     cp.destroy ()
